import hashlib
from tqdm import tqdm
from pathlib import Path
from typing import Union, IO, AnyStr
from abc import abstractmethod, abstractproperty


class Provider:
    @abstractmethod
    def __init__(self, config):
        raise NotImplementedError

    @abstractmethod
    def download(
        self, key: str, file_path: Union[str, Path] = None, buffer: IO[AnyStr] = None
    ):
        raise NotImplementedError

    @abstractmethod
    def upload(
        self, key: str, file_path: Union[str, Path] = None, buffer: IO[AnyStr] = None
    ):
        raise NotImplementedError

    @abstractproperty
    def keys(self):
        raise NotImplementedError

    @abstractmethod
    def check_valid(self, key: str, file_path: Union[str, Path]):
        raise NotImplementedError

    @abstractmethod
    def file_size(self, key: str):
        raise NotImplementedError

    @abstractmethod
    def delete(self, key: str):
        raise NotImplementedError

    def md5(self, file_path: Union[str, Path] = None, buffer: IO[AnyStr] = None):
        md5_hash = hashlib.md5()

        if file_path is not None:
            buffer = Path(file_path).open("rb")

        msg = f"`buffer` and `file_path` cannot be `None` simultaneously."
        assert buffer is not None, msg

        content = buffer.read()
        md5_hash.update(content)
        return md5_hash.hexdigest()


class Progress:
    def __init__(self, size: int, key: str):
        self.size = size
        self.progress = 0
        self.progress_bar = tqdm(desc=f"Downloading {key}", total=size, leave=False)

    def __call__(self, bytes: int):
        self.progress += bytes
        self.progress_bar.update(bytes)

    def close(self):
        self.progress_bar.update(self.size - self.progress)
        self.progress_bar.close()

    def __exit__(self):
        self.progress_bar.close()



import argparse
from daman import daman as dm


def delete_command():
    parser = argparse.ArgumentParser(description="Deletes file.")
    parser.add_argument(
        "--key", type=str, help="key of the file to delete", required=True
    )
    parser.add_argument(
        "--remote",
        action="store_true",
        help="When provided, deletes the requested file on the cloud service as well.",
    )
    args = parser.parse_args()

    dm.delete(
        key=args.key, local=True, remote=args.remote,
    )



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fr:
    requirements = fr.read().splitlines()

setup(
    name="daman",
    author="Sofiane Mahiou",
    author_email="mahiou.sofiane@gmail.com",
    description=(
        "A simple package that allows for the direct use of data"
        " stored and managed on a cloud storage service"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sofiane87/daman",
    python_requires=">=3.6.0",
    version="0.1.5",
    packages=find_packages("."),
    package_dir={"": "."},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "dm_configure = daman.commands:configure_command",
            "dm_aws = daman.commands:configure_aws_command",
            "dm_pull = daman.commands:download_command",
            "dm_delete = daman.commands:delete_command",
            "dm_clear = daman.commands:clear_command",
        ]
    },
    project_urls={"Bug Tracker": "https://github.com/sofiane87/daman/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
)



import argparse
from daman import daman as dm


def download_command():
    parser = argparse.ArgumentParser(description="Downloads file")
    parser.add_argument(
        "--key", type=str, help="key of the file to delete", required=True
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="When provided forces the download even when the file is already available.",
    )
    parser.add_argument(
        "--persist",
        action="store_true",
        help="When provided ensures that the downloaded file is always kept on disc on manually deleted.",
    )
    args = parser.parse_args()

    dm.pull(key=args.key, force=args.force, persist=args.persist)



import os
import joblib
import psutil
from logging import getLogger
from io import BytesIO
from pathlib import Path
from os.path import getsize
from datetime import datetime

from daman.utils import dir_size
from configparser import ConfigParser

from daman.configure import CONFIG_DIR
from daman.services import PROVIDERS
from daman.data import DataRegistery


logger = getLogger(__name__)


class DataManager:
    @property
    def config(self):
        config = ConfigParser()
        config.read(CONFIG_DIR / "config")
        return config

    @property
    def data_folder(self):
        config = self.config
        d_folder = Path(config["local"]["data_dir"])
        if not d_folder.exists():
            d_folder.mkdir(parents=True, exist_ok=True)
        return d_folder

    @property
    def registery(self):
        return DataRegistery()

    @property
    def service(self):
        req_service = self.config["service"]["service"]
        if req_service in PROVIDERS:
            return PROVIDERS[req_service](config=self.config)
        else:
            raise KeyError(
                f"service `{req_service}` requested is not available among provided services."
            )

    def pull(
        self,
        key: str,
        force: bool = False,
        persist: bool = None,
        memory_only: bool = False,
    ):
        msg = f"key '{key}' not found"
        assert key in self.service.keys, msg
        if key in self.registery and not force:
            if persist is not None:
                self.registery[key]["persist"] = persist
            self.registery[key]["used"] += 1
            self.registery[key]["last_used"] = datetime.now()
            file_path = Path(self.registery[key]["path"])
            is_valid = self.service.check_valid(key=key, file_path=file_path)
            if is_valid:
                logger.info(f"data `{key}` available locally.")
                obj = joblib.load(file_path)
                return obj["data"], obj["meta"]
            else:
                logger.warning(
                    f"local '{key}' file doesn't match remote. Data will be downloaded again."
                )

        # download file
        if persist is None:
            persist = False

        if memory_only:
            buffer = BytesIO()
            self.service.download(key=key, buffer=buffer)
            obj = joblib.load(buffer)
            return obj["data"], obj["meta"]
        else:
            logger.info(f"downloading `{key}` file.")
            self.clear_disc(key=key)
            file_path = (Path(self.data_folder) / key).resolve()
            self.service.download(key=key, file_path=file_path)
            self.registery[key] = {
                "path": str(file_path),
                "used": 1,
                "created_at": datetime.now(),
                "last_used": datetime.now(),
                "size": getsize(str(file_path)),
                "persist": persist,
            }
            obj = joblib.load(file_path)
            return obj["data"], obj["meta"]

    def push(
        self,
        obj: object,
        key: str,
        meta: object = None,
        force: bool = False,
        persist: bool = False,
    ):
        obj = {"data": obj, "meta": meta}
        if key in self.service.keys:
            if force:
                logger.warning(f"`{key}` already in use and will be over-written.")
            else:
                err_msg = f"`{key}` already in use. Please choose a different key or use --force to enforce key override."
                logger.error(err_msg)
                raise KeyError(err_msg)

        logger.info(f"storing `{key}` data locally.")
        # store locally
        file_path = (Path(self.data_folder) / key).resolve()
        # store locally
        joblib.dump(obj, file_path)

        logger.info(f"uploading {key} to cloud service.")
        # upload to cloud
        self.service.upload(key=key, file_path=file_path)

        # update registery
        self.registery[key] = {
            "path": str(file_path),
            "used": 1,
            "created_at": datetime.now(),
            "last_used": datetime.now(),
            "size": getsize(str(file_path)),
            "persist": persist,
        }

    def delete(self, key: str, local: bool = True, remote: bool = False):
        if local:
            logger.info(f"deleting `{key}` data locally.")
            item = self.registery[key]
            # delete file
            del self.registery[key]
            # delete local file
            os.remove(item["path"])
        if remote:
            logger.info(f"deleting `{key}` data on cloud service.")
            # delete remote file
            self.service.delete(key=key)

    def list_data(self, local_only: bool = False):
        return [
            {
                "key": key,
                "local": key in self.registery,
                "size": self.service.file_size(key) // 2 ** 20,
                "remote": key in self.service.keys,
            }
            for key in (
                self.registery.keys
                if local_only
                else set(self.service.keys) | set(self.registery.keys)
            )
        ]

    def available_disc(self):
        if self.config["local"]["allocated_space"] is None:
            free_space = (
                psutil.disk_usage(self.data_folder).free() // 2 ** 20
            )  # disc space in megabytes
        else:
            free_space = int(self.config["local"]["allocated_space"]) - dir_size(
                self.data_folder
            )
        return free_space

    def clear_disc(
        self, key: str = None, space: int = None, ignore_persist: bool = False
    ):
        if key is not None:
            file_size = self.service.file_size(key=key) // 2 ** 20
        else:
            file_size = space

        while self.available_disc() < file_size:
            logger.info(
                f"deleting `{file_size - self.available_disc()}` data on cloud service."
            )
            items = sorted(
                self.registery.items,
                key=lambda x: (x[1]["persist"], x[1]["last_used"], x[1]["used"]),
            )

            if len(items) > 0:
                del_key, del_item = items[0]
                if not del_item["persist"] or ignore_persist:
                    self.delete(key=del_key, local=True, remote=False)
                    continue

            raise IOError(
                f"Not enough space available. Only {self.available_disc()} MB available but {file_size} MB required. No additional file can be deleted."
            )



from typing import Union
from pathlib import Path
from os.path import getsize

unit_map = {None: 2 ** 0, "K": 2 ** 10, "M": 2 ** 20, "G": 2 ** 30}


def dir_size(dir: Union[str, Path], unit="M"):
    msg = f"unit '{unit}' not recognized. Allowed units: [{list(unit_map.keys())}]"
    assert unit in unit_map, msg

    dir_size = 0
    for file in Path(dir).rglob("*"):
        if file.is_file():
            dir_size += getsize(dir)

    return dir_size // unit_map[unit]
